"""Calibration of predicted probabilities."""
from __future__ import division

import numpy as np
import sklearn
import random
import matplotlib.pyplot as plt
from scipy.stats import binom

try:
    from sklearn.model_selection import StratifiedKFold, KFold
except:
    from sklearn.cross_validation import StratifiedKFold, KFold

from .utils import _gca


def _natural_cubic_spline_basis_expansion(xpts, knots):
    num_knots = len(knots)
    num_pts = len(xpts)
    outmat = np.zeros((num_pts,num_knots))
    outmat[:, 0] = np.ones(num_pts)
    outmat[:, 1] = xpts

    def make_func_H(k):
        def make_func_d(k):
            def func_d(x):
                denom = knots[-1] - knots[k-1]
                numer = np.maximum(x-knots[k-1], np.zeros(len(x))) ** 3 - np.maximum(x-knots[-1], np.zeros(len(x))) ** 3
                return numer/denom
            return func_d

        def func_H(x):
            d_fun_k = make_func_d(k)
            d_fun_Km1 = make_func_d(num_knots-1)
            return d_fun_k(x) -  d_fun_Km1(x)
        return func_H
    for i in range(1, num_knots-1):
        curr_H_fun = make_func_H(i)
        outmat[:, i+1] = curr_H_fun(xpts)
    return outmat


def prob_calibration_function(truthvec, scorevec, reg_param_vec='default', knots='sample',
                              method='logistic', force_prob=True, eps=1e-15, max_knots=200,
                              transform_fn='none', random_state=942, verbose=False, cv_folds=5,
                              unity_prior_weight=1, unity_prior_gridsize=20):
    """This function takes an uncalibrated set of scores and the true 0/1 values and returns a calibration function.

    This calibration function can then be applied to other scores from the same model and will return an accurate probability
    based on the data it has seen.  For best results, the calibration should be done on a separate validation set (not used
    to train the model).

    Parameters
    ----------
    truthvec : A numpy array containing the true values that is the target of the calibration.  For binary
     classification these are typically your 0/1 values.

    scorevec : A numpy array containing the scores that are not appropriate to be used as probabilities.  These do not
     necessarily need to be between 0 and 1, though that is the typical usage.

    reg_param_vec:  The vector of C-values (if method = 'logistic') or alpha values (if method = 'ridge') that the calibration should
      search across.  If reg_param_vec = 'default' (which is the default) then it picks a reasonable set of values to search across.

    knots: Default is 'sample', which means it will randomly pick a subset of size max_knots from the unique values in scorevec (while always
        keeping the largest and smallest value).  If knots='all' it will use all unique values of scorevec as knots.  This may yield a
        better calibration, but will be slower.

    method : 'logistic' or 'ridge'
        The default is 'logistic', which is best if you plan to use log-loss as your metric.  You can also use "ridge" which may
        be better if Brier Score is your metic of interest.  However, "ridge" may be less stable and robust, especially when used
        for probabilities.

    force_prob: This is ignored for method = 'logistic'.  For method = 'ridge', if set to True (the default), it will ensure that the
        values coming out of the calibration are between eps and 1-eps.

    eps: default is 1e-15.  Applies only if force_prob = True and method = 'ridge'.  See force_prob above.

    max_knots:  The number of knots to use when knots='sample'.  See knots above.

    random_state: default is 942 (a particular value to ensure consistency when running multiple times).  User can supply a different value
        if they want a different random seed.

    Returns
    ---------------------

    A function object which takes a numpy array (or a single number) and returns the output of the calculated calibration function.

    """
    from sklearn import linear_model
    from sklearn.metrics import log_loss, make_scorer

    if (unity_prior_weight>0):
        scorevec_coda, truthvec_coda = create_yeqx_bias_vectors(unity_prior_gridsize)
        coda_wt = unity_prior_weight/unity_prior_gridsize
        weightvec = np.concatenate((np.ones(len(scorevec)), coda_wt * np.ones(len(scorevec_coda))))
        scorevec = np.concatenate((scorevec, scorevec_coda))
        truthvec = np.concatenate((truthvec, truthvec_coda))

    if transform_fn != 'none':
        scorevec = transform_fn(scorevec)

    knot_vec = np.unique(scorevec)
    if (knots == 'sample'):
        num_unique = len(knot_vec)
        if (num_unique > max_knots):
            smallest_knot, biggest_knot = knot_vec[0], knot_vec[-1]
            inter_knot_vec = knot_vec[1:-1]
            random.seed(random_state)
            random.shuffle(inter_knot_vec)
            reduced_knot_vec = inter_knot_vec[:(max_knots-2)]
            reduced_knot_vec = np.concatenate((reduced_knot_vec, [smallest_knot, biggest_knot]))
            reduced_knot_vec = np.concatenate((reduced_knot_vec, np.linspace(0, 1, 21)))
            if (unity_prior_weight>0):
                reduced_knot_vec = np.concatenate((reduced_knot_vec, scorevec_coda))
            knot_vec = np.unique(reduced_knot_vec)
        if verbose:
            print("Originally there were {} knots.  Reducing to {} while preserving first and last knot.".format(num_unique, len(knot_vec)))
    X_mat = _natural_cubic_spline_basis_expansion(scorevec, knot_vec)

    if (method == 'logistic'):
        if ((type(reg_param_vec) == str) and (reg_param_vec == 'default')):
            reg_param_vec = 10**np.linspace(-7, 5, 61)
        if verbose:
            print("Trying {} values of C between {} and {}".format(len(reg_param_vec), np.min(reg_param_vec), np.max(reg_param_vec)))
        reg = linear_model.LogisticRegressionCV(Cs=reg_param_vec, cv=StratifiedKFold(cv_folds, shuffle=True),
                                                scoring=make_scorer(log_loss, needs_proba=True, greater_is_better=False))
        if (unity_prior_weight>0):
            reg.fit(X_mat, truthvec, weightvec)
        else:
            reg.fit(X_mat, truthvec)
        if verbose:
            print("Best value found C = {}".format(reg.C_))

    if (method == 'ridge'):
        if ((type(reg_param_vec) == str) and (reg_param_vec == 'default')):
            reg_param_vec = 10**np.linspace(-7, 7, 71)
        if verbose:
            print("Trying {} values of alpha between {} and {}".format(len(reg_param_vec), np.min(reg_param_vec),np.max(reg_param_vec)))
        reg = linear_model.RidgeCV(alphas=reg_param_vec, cv=KFold(cv_folds, shuffle=True), scoring=make_scorer(mean_squared_error_trunc,needs_proba=False, greater_is_better=False))
        reg.fit(X_mat, truthvec)
        if verbose:
            print("Best value found alpha = {}".format(reg.alpha_))

    def calibrate_scores(new_scores):
        new_scores = np.maximum(new_scores,knot_vec[0]*np.ones(len(new_scores)))
        new_scores = np.minimum(new_scores,knot_vec[-1]*np.ones(len(new_scores)))
        if transform_fn != 'none':
            new_scores = transform_fn(new_scores)
        basis_exp = _natural_cubic_spline_basis_expansion(new_scores,knot_vec)
        if (method == 'logistic'):
            outvec = reg.predict_proba(basis_exp)[:,1]
        if (method == 'ridge'):
            outvec = reg.predict(basis_exp)
            if force_prob:
                outvec = np.where(outvec < eps, eps, outvec)
                outvec = np.where(outvec > 1-eps, 1-eps, outvec)
        return outvec

    return calibrate_scores


def mean_squared_error_trunc(y_true, y_pred,eps=1e-15):
    y_pred = np.where(y_pred<eps,eps,y_pred)
    y_pred = np.where(y_pred>1-eps,1-eps,y_pred)
    return np.average((y_true-y_pred)**2)


def prob_calibration_function_multiclass(truthvec, scoremat, verbose=False, **kwargs):
    """This function takes an uncalibrated set of scores and the true 0/1 values and returns a calibration function.

    This calibration function can then be applied to other scores from the same model and will return an accurate probability
    based on the data it has seen.  For best results, the calibration should be done on a separate validation set (not used
    to train the model).

    Parameters
    ----------
    truthvec : A numpy array containing the true values that is the target of the calibration.  For binary
     classification these are typically your 0/1 values.

    scorevec : A numpy array containing the scores that are not appropriate to be used as probabilities.  These do not
     necessarily need to be between 0 and 1, though that is the typical usage.

    reg_param_vec:  The vector of C-values (if method = 'logistic') or alpha values (if method = 'ridge') that the calibration should
      search across.  If reg_param_vec = 'default' (which is the default) then it picks a reasonable set of values to search across.

    knots: Default is 'sample', which means it will randomly pick a subset of size max_knots from the unique values in scorevec (while always
        keeping the largest and smallest value).  If knots='all' it will use all unique values of scorevec as knots.  This may yield a
        better calibration, but will be slower.

    method : 'logistic' or 'ridge'
        The default is 'logistic', which is best if you plan to use log-loss as your metric.  You can also use "ridge" which may
        be better if Brier Score is your metic of interest.  However, "ridge" may be less stable and robust, especially when used
        for probabilities.

    force_prob: This is ignored for method = 'logistic'.  For method = 'ridge', if set to True (the default), it will ensure that the
        values coming out of the calibration are between eps and 1-eps.

    eps: default is 1e-15.  Applies only if force_prob = True and method = 'ridge'.  See force_prob above.

    max_knots:  The number of knots to use when knots='sample'.  See knots above.

    random_state: default is 942 (a particular value to ensure consistency when running multiple times).  User can supply a different value
        if they want a different random seed.

    Returns
    ---------------------

    A function object which takes a numpy array (or a single number) and returns the output of the calculated calibration function.

    """
    from sklearn import linear_model
    from sklearn.metrics import log_loss, make_scorer

    num_classes = scoremat.shape[1]
    function_list = []
    for i in range(num_classes):
        scorevec = scoremat[:,i]
        curr_truthvec = (truthvec==i).astype(int)
        function_list.append(prob_calibration_function(curr_truthvec,scorevec,verbose=verbose,**kwargs))

    def calibrate_scores_multiclass(new_scoremat):
        a,b = new_scoremat.shape
        pre_probmat = np.zeros((a,b))
        for i in range(num_classes):
            pre_probmat[:,i] = function_list[i](new_scoremat[:,i])
        probmat = (pre_probmat.T/np.sum(pre_probmat,axis=1)).T
        #if (not extrapolate):
        #    new_scores = np.maximum(new_scores,knot_vec[0]*np.ones(len(new_scores)))
        #    new_scores = np.minimum(new_scores,knot_vec[-1]*np.ones(len(new_scores)))
        return probmat
    return calibrate_scores_multiclass, function_list

def plot_prob_calibration(calib_fn, show_baseline=True, ax=None, **kwargs):
    if ax is None:
        ax = _gca()
        fig = ax.get_figure()
    ax.plot(np.linspace(0,1,100),calib_fn(np.linspace(0,1,100)),**kwargs)
    if show_baseline:
        ax.plot(np.linspace(0,1,100),(np.linspace(0,1,100)),'k--')
    ax.axis([-0.1,1.1,-0.1,1.1])

def plot_reliability_diagram(y,x,bins=np.linspace(0,1,21),size_points=False, show_baseline=True,
                                error_bars=True, error_bar_alpha = .05, marker='+',c='red', **kwargs):
    # if ax is None:
    #     ax = plt.gca()
    #     fig = ax.get_figure()
    digitized_x = np.digitize(x, bins)
    mean_count_array = np.array([[np.mean(y[digitized_x == i]),len(y[digitized_x == i]),np.mean(x[digitized_x==i])] for i in np.unique(digitized_x)])
    x_pts_to_graph = mean_count_array[:,2]
    y_pts_to_graph = mean_count_array[:,0]
    pt_sizes = mean_count_array[:,1]
    plt.subplot(1,2,1)

    if show_baseline:
        plt.plot(np.linspace(0,1,100),(np.linspace(0,1,100)),'k--')
#        ax.loglog(np.linspace(0,1,100),(np.linspace(0,1,100)),'k--')
    for i in range(len(y_pts_to_graph)):
        if size_points:
            plt.scatter(x_pts_to_graph,y_pts_to_graph,s=pt_sizes,marker=marker,c=c, **kwargs)
        else:
            plt.scatter(x_pts_to_graph,y_pts_to_graph, c=c, **kwargs)
    plt.axis([-0.1,1.1,-0.1,1.1])
    
    if error_bars:
        yerr_mat = binom.interval(1-error_bar_alpha,pt_sizes,x_pts_to_graph)/pt_sizes - x_pts_to_graph
        yerr_mat[0,:] = -yerr_mat[0,:]
        plt.errorbar(x_pts_to_graph, x_pts_to_graph, yerr=yerr_mat, capsize=5)
    plt.subplot(1,2,2)
    plt.hist(x,bins=bins)

    return(x_pts_to_graph,y_pts_to_graph,pt_sizes)

def compact_logit(x, eps=.00001):
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in log")
        warnings.filterwarnings("ignore", message="invalid value encountered in multiply")
        return np.nansum(((x<=eps)*x, (x>=(1-eps))*x, ((x>eps)&(x<(1-eps)))*((1-2*eps)*(np.log(x/(1-x)))/(2*np.log((1-eps)/eps))+.5)),axis=0)

def inverse_compact_logit(x, eps=.00001):
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in log")
        warnings.filterwarnings("ignore", message="invalid value encountered in multiply")
        return np.nansum(((x<=eps)*x, (x>=(1-eps))*x,
                          ((x>eps)&(x<(1-eps)))*
                          (1/(1+np.exp(-(x-.5)*((2/(1-2*eps))*np.log((1-eps)/eps)))))),axis=0)


def create_yeqx_bias_vectors(gridsize=10):
    scorevec_coda = np.sort(np.tile(np.arange(gridsize + 1)/gridsize, reps = (gridsize)))
    truthvec_coda = np.array([])
    for i in range(gridsize + 1):
        added_bit = np.concatenate((np.zeros(gridsize - i), np.ones(i)))
        truthvec_coda = np.concatenate((truthvec_coda, added_bit))
    return scorevec_coda, truthvec_coda



import numpy as np

class graph_undirected(object):
    """This is a class to handle undirected graphs.  Still very much a work in progress.
    Defines a graph by a set of vertices and a set of "frozensets" representing the edges.
    """

    def __init__(self, edges, vertices={}):
        
        if vertices == {}:
            self.vertices = set([x for sublist in list(edges) for x in list(sublist) ])
        else:
            self.vertices = vertices
        new_edges = set()
        for edge in edges:
            new_edge = frozenset(edge)
            if len(new_edge)>1:
                new_edges.add(new_edge)
        self.edges = set(new_edges)


    def adjacent_edges(self, target_vertex):
        return set([x for x in self.edges if target_vertex in x])

    def adjacent_vertices(self, target_vertex):
        neighbors_and_self = set([x for sublist in self.adjacent_edges(target_vertex) for x in sublist])
        return set(neighbors_and_self)-set([target_vertex])

    def adjacent_vertices_to_set(self, target_vertex_set):
        templist = [list(self.adjacent_vertices(x)) for x in target_vertex_set]
        neighbors_and_self = [x for sublist in templist for x in sublist]
        return set(neighbors_and_self)-target_vertex_set
    
    def vertex_degree(self, target_vertex):
        return len(self.adjacent_vertices(target_vertex))

    def contract_edge(self, edge):
        return contract_edge(self, edge)

    def delete_vertex(self, vertex):
        return delete_vertex(self, vertex)

    def delete_vertices(self, vertex_set):
        return delete_vertices(self, vertex_set)

    def get_induced_subgraph(self, vertex_set):
        return get_induced_subgraph(self, vertex_set)


    def enumerate_all_partitions(self, max_size=0, memory_save=False, verbose=False):
        if(len(self.vertices)==0):
            return []
        if((len(self.vertices)>=1) and (len(self.vertices)<=2)):
            return [frozenset([list(self.vertices)[0]])]
        if max_size==0:
            max_size=int(np.floor(len(self.vertices)/2))
        good_partitions = []
        failed_partitions = []
        good_partitions.append(set())
        failed_partitions.append(set())
        num_good_partition_list = []
        num_failed_partition_list = []
        if verbose:
            print('Evaluating components of size 1')
        for vert in self.vertices:
            if is_connected(delete_vertex(self, vert)):
                good_partitions[0].add(frozenset({vert}))
            else:
                failed_partitions[0].add(frozenset({vert}))
        num_good_partition_list.append(len(good_partitions[0]))
        num_failed_partition_list.append(len(failed_partitions[0]))
        if verbose:
            print('num_good_partitions of comp_size 1 = {}'.format(num_good_partition_list[0]))
            print('num_failed_partitions of comp_size 1 = {}'.format(num_failed_partition_list[0]))
            print('Evaluating components of size 2')
        good_partitions.append(set())
        failed_partitions.append(set())
        for edge in self.edges:
            if is_connected(delete_vertices(self, edge)):
                good_partitions[1].add(edge)
            else:
                failed_partitions[1].add(edge)
        num_good_partition_list.append(len(good_partitions[1]))
        num_failed_partition_list.append(len(failed_partitions[1]))
        if verbose:
            print('num_good_partitions of comp_size 2 = {}'.format(num_good_partition_list[1]))
            print('num_failed_partitions of comp_size 2 = {}'.format(num_failed_partition_list[1]))
            print('num_total_good_partitions of comp_size<=2 is {}'.format(np.sum(num_good_partition_list)))
            print('num_total_failed_partitions of comp_size <=2 is {}'.format(np.sum(num_failed_partition_list)))

        
        for comp_size in range(3, max_size+1):
            good_partitions.append(set())
            failed_partitions.append(set())

            if verbose:
                print('Evaluating components of size {}'.format(comp_size))
            base_components = good_partitions[comp_size-2]
            for base_comp in base_components:
                neighbors_to_add = self.adjacent_vertices_to_set(base_comp)
                for neighbor in neighbors_to_add:
                    new_comp = set(base_comp)
                    new_comp.add(neighbor)
                    new_comp = frozenset(new_comp)
                    if ((not new_comp in good_partitions[comp_size-1]) and (not new_comp in failed_partitions[comp_size-1])):
                        if is_connected(delete_vertices(self,new_comp)):
                            good_partitions[comp_size-1].add(new_comp)
                        else:
                            failed_partitions[comp_size-1].add(new_comp)
            num_good_partition_list.append(len(good_partitions[comp_size-1]))
            num_failed_partition_list.append(len(failed_partitions[comp_size-1]))
            if memory_save:
                good_partitions[comp_size-2]={}
                failed_partitions[comp_size-2]={}
                
            if verbose:
                print('num_good_partitions of comp_size {} = {}'.format(comp_size,num_good_partition_list[comp_size-1]))
                print('num_failed_partitions of comp_size {} = {}'.format(comp_size,num_failed_partition_list[comp_size-1]))
                print('num_total_good_partitions of comp_size<={} is {}'.format(comp_size, np.sum(num_good_partition_list)))
                print('num_total_failed_partitions of comp_size <={} is {}'.format(comp_size, np.sum(num_failed_partition_list)))
             
        good_partitions = [k for templist in good_partitions for k in templist]
        return good_partitions

    
def contract_edge(graph, edge):
    edge_alph = list(edge)
    edge_alph.sort()
    contracted_vertex = '_'.join((edge_alph))
    #new_vertices = (set(graph.vertices) - set(edge)).union({contracted_vertex})
    new_edges = [[contracted_vertex if y==edge_alph[0] or y==edge_alph[1] else y for y in this_edge] 
                 if edge_alph[0] in this_edge  or edge_alph[1] in this_edge else this_edge for this_edge in graph.edges]
    return graph_undirected(new_edges)

def delete_vertex(graph, vertex):
    new_edges = set([edge for edge in graph.edges if vertex not in edge])
    new_vertices = graph.vertices - {vertex}
    return graph_undirected(new_edges, new_vertices)

def delete_vertices(graph, vertex_set):
    new_edges = set([edge for edge in graph.edges if not vertex_set.intersection(edge)])
    new_vertices = graph.vertices - vertex_set
    return graph_undirected(new_edges, new_vertices)

def get_induced_subgraph(graph, vertex_set):
    vertex_set = set(vertex_set)
    new_edges = set([edge for edge in graph.edges if edge <= vertex_set])
    new_vertices = vertex_set
    return graph_undirected(new_edges, new_vertices)


def is_connected(graph):
    initial_vertex = list(graph.vertices)[0]
    visited_vertices = [initial_vertex]
    unexplored_vertices = list(graph.adjacent_vertices(initial_vertex))
    while unexplored_vertices:
        curr_vertex = unexplored_vertices.pop()
        visited_vertices.append(curr_vertex)
        new_vertices = graph.adjacent_vertices(curr_vertex)
        unexplored_vertices = list(set(unexplored_vertices).union(new_vertices) - set(visited_vertices))
    return len(set(visited_vertices)) == len(set(graph.vertices))

def num_connected_comp(graph):
    initial_vertex = list(graph.vertices)[0]
    visited_vertices = [initial_vertex]
    unexplored_vertices = list(graph.adjacent_vertices(initial_vertex))
    while unexplored_vertices:
        curr_vertex = unexplored_vertices.pop(0)
        visited_vertices.append(curr_vertex)
        new_vertices = graph.adjacent_vertices(curr_vertex)
        unexplored_vertices = list(set(unexplored_vertices).union(new_vertices) - set(visited_vertices))
    if len(set(visited_vertices)) == len(set(graph.vertices)):
        return 1
    else:
        remainder_vertices = list(set(graph.vertices)-set(visited_vertices))
        remainder_edges = [edge for edge in graph.edges if edge.issubset(set(remainder_vertices))]
        return 1 + num_connected_comp(graph_undirected(remainder_edges, remainder_vertices))
    
def connected_comp_list(graph):
    initial_vertex = list(graph.vertices)[0]
    visited_vertices = [initial_vertex]
    unexplored_vertices = list(graph.adjacent_vertices(initial_vertex))
    while unexplored_vertices:
        curr_vertex = unexplored_vertices.pop(0)
        visited_vertices.append(curr_vertex)
        new_vertices = graph.adjacent_vertices(curr_vertex)
        unexplored_vertices = list(set(unexplored_vertices).union(new_vertices) - set(visited_vertices))
    if len(set(visited_vertices)) == len(set(graph.vertices)):
        return [graph]
    else:
        cc_vertices = set(visited_vertices)
        cc_edges = [edge for edge in graph.edges if edge.issubset(set(visited_vertices))]
        cc_graph = graph_undirected(cc_edges, cc_vertices)
        remainder_vertices = list(set(graph.vertices)-set(visited_vertices))
        remainder_edges = [edge for edge in graph.edges if edge.issubset(set(remainder_vertices))]
        return [cc_graph] + connected_comp_list(graph_undirected(remainder_edges, remainder_vertices))
    
def get_all_distances_from_vertex(graph, start_vertex):
    vertex_path_dist_dict={}
    vertex_path_dist_dict[start_vertex] = 0
    unexplored_vertices = list(graph.adjacent_vertices(start_vertex))
    for vert in unexplored_vertices:
        vertex_path_dist_dict[vert]=1
    visited_vertices = [start_vertex]
    
    while unexplored_vertices and (len(vertex_path_dist_dict.keys())<len(graph.vertices)):
        curr_vertex = unexplored_vertices.pop()
        curr_dist = vertex_path_dist_dict[curr_vertex]
        visited_vertices.append(curr_vertex)
        curr_neighbors = graph.adjacent_vertices(curr_vertex)
        new_vertices = curr_neighbors - set(vertex_path_dist_dict.keys())
        for vert in new_vertices:
            vertex_path_dist_dict[vert]=curr_dist+1
        unexplored_vertices = list(new_vertices) + unexplored_vertices
    return vertex_path_dist_dict

def separate_by_two_vertices(graph, vert_1, vert_2):
    dict_1 = get_all_distances_from_vertex(graph, vert_1)
    dict_2 = get_all_distances_from_vertex(graph, vert_2)
    comp_2 = set([vert for vert in dict_2.keys() if dict_2[vert]<dict_1[vert]])
    comp_1 = graph.vertices - comp_2
    return comp_1, comp_2
